
-- | A custom envelope widget for wxHaskell.
-- The word \"envelope\" is understood as in ADSR.
-- The envelope is built from linear or cubic bezier segments.
--
-- The code is very ugly and hackish. You have been warned.
 
module Graphics.UI.WX.Controls.EnvelopeWidget 
  ( SegmentType(..)
  , EnvSegment(..)
  , Envelope(..)
  , Envelope'
  , calcEnvelopeSegment
  , calcEnvelope
  , makeEnvelopeWidget
  ) where

import Control.Monad
import Control.Concurrent.MVar

import Graphics.UI.WX     hiding (when)
import Graphics.UI.WXCore hiding (when)

import Data.Envelope

-------------------------------------------------------

clamp :: Ord a => (a,a) -> a -> a
clamp (minv,maxv) x = max (min x maxv) minv

adjustMVar :: MVar a -> (a -> a) -> IO ()
adjustMVar mv f = do
  x <- takeMVar mv
  putMVar mv (f x)

-------------------------------------------------------

data PointType = Endpoint | Control

data ControlPoint = ControlPoint
  { cp_win     :: Panel ()
  , cp_parent  :: Panel ()
  , cp_env     :: EnvelopeWidget
  , cp_type    :: PointType
  , cp_signed  :: Bool
  , cp_index   :: (Int,Int) 
  , cp_enabled :: MVar Bool
  , cp_pos     :: MVar (Double,Double)
  , cp_state   :: MVar ControlPointState
  }

data ControlPointState 
  = Up 
  | Down 
      { cps_startPos :: (Int,Int) 
      , cps_oldPos :: (Double,Double) 
      }

getPos :: ControlPoint -> IO (Double,Double)
getPos cp = readMVar (cp_pos cp)

setXPos :: ControlPoint -> Double -> IO ()
setXPos cp x = do 
  (_,y) <- readMVar (cp_pos cp)
  setPos cp (x,y)
  
setPos :: ControlPoint -> (Double,Double) -> IO ()
setPos cp xy = do
  swapMVar (cp_pos cp) xy
  updatePos cp
  
  let env = cp_env cp
  -- wxwidgets can't draw the same line twice with different 
  -- colors (anti-aliasing at work...)
  -- thus we need a fat background-colored line to clear the thin
  -- non-background-colored line... 
  drawBezier1 env backgroundColor backgroundColor (2+bezierWidth) 
  case cp_type cp of
    Endpoint -> adjustMVar (_envelope $ cp_env cp) $ liftEnv (updateNthEndpoint (fst $ cp_index cp) xy)
    Control  -> adjustMVar (_envelope $ cp_env cp) $ liftEnv (updateNthControl  (cp_index cp) (snd xy))
  drawBezier1 env bezierColor tangentColor bezierWidth
  repaintAllControls env

updatePos :: ControlPoint -> IO () 
updatePos cp = do 
  (x,y) <- readMVar (cp_pos cp)
  Rect _ _ w h <- windowGetRect (cp_parent cp)
  let (a,b) = case cp_signed cp of 
        False -> ( round (x * fromIntegral w) - 5 
                 , round (y * fromIntegral h) - 5 
                 )     
        True  -> ( round (          x  * fromIntegral w) - 5
                 , round ( 0.5 * (1+y) * fromIntegral h) - 5 
                 )     
  set (cp_win cp) [ position := Point a b ]
  
makeControlPoint :: EnvelopeWidget -> PointType -> Bool -> (Int,Int) -> Bool -> [Prop ControlPoint] -> IO ControlPoint
makeControlPoint ew ptype signed index enabled props = do 
  let parent = _window ew
  win <- panelCreate parent 0 (Rect 0 0 12 12) 0 
  pos <- newMVar (0,0) 
  enb <- newMVar enabled
  state <- newMVar Up
  let cp = ControlPoint win parent ew ptype signed index enb pos state
  set win 
    [ on paint := paintControlPoint win ptype  
    , on mouse := mouseControlPoint cp 
    , visible  := enabled
    ]
  set cp props
  return cp
  
instance Able ControlPoint where
  enabled = newAttr "able" getter setter where
    getter cp = readMVar (cp_enabled cp)
    setter cp x = do
      swapMVar (cp_enabled cp) x 
      set (cp_win cp) [ visible := x ]  
  
paintControlPoint :: Panel () -> PointType -> DC () -> Rect -> IO ()   
paintControlPoint win ptype dc rect = do
  dcClear dc
  let 
    col = case ptype of
      Endpoint -> colorRGB 255 255 0
      Control  -> colorRGB 0   255 0     
  dcWithPenStyle dc (penColored col 2) $ do
    dcDrawCircle dc (Point 6 6) 5  
 
mouseControlPoint :: ControlPoint -> EventMouse -> IO ()   
mouseControlPoint cp event = do
  let state  = cp_state  cp
      win    = cp_win    cp
      pos    = cp_pos    cp
      parent = cp_parent cp
      signed = cp_signed cp
      ew     = cp_env    cp
  Rect l t w h <- windowGetRect parent
  s <- takeMVar state
  s' <- case s of
    Up -> case event of
      MouseLeftDown (Point x y) mod -> do
        oldpos <- readMVar pos
        windowCaptureMouse win
        Point a0 b0 <- get win position
        return $ Down (a0+x,b0+y) oldpos 
      _ -> return s
    Down (x0,y0) pq0@(p0,q0) -> case event of
      MouseLeftDrag (Point x' y') mod -> do
        Point a0 b0 <- get win position
        let (x,y) = (a0+x',b0+y')

        -- logStatus $ show ((l,t,w,h),(x,y))
        
        env@(Envelope n env') <- readMVar (_envelope ew)
        
        let --p = p0 -- clamp (0,1) $ p0 + fromIntegral (x-x0) / fromIntegral w
            i = fst (cp_index cp)
            
            p = case cp_type cp of
              Control -> p0
              Endpoint -> guarded i where
                p' = p0 + fromIntegral (x-x0) / fromIntegral w 
                guarded i 
                  | i == 0     = p0
                  | i >= n     = p0
                  | otherwise  = if (b-a) <= 2*limit then p0 else clamp (a+limit,b-limit) p'
                  where 
                    limit = 0.075
                    a = nthTimePoint (i-1) env'
                    b = nthTimePoint (i+1) env'
                
            q = if signed 
              then clamp (-1,1) $ q0 + 2*fromIntegral (y-y0) / fromIntegral h
              else clamp ( 0,1) $ q0 +   fromIntegral (y-y0) / fromIntegral h

        setPos cp (p,q)
        Control.Monad.when (p/=p0) $ do
          let a = nthTimePoint (i-1) env'
              b = nthTimePoint (i+1) env'
          controls <- readMVar (_controls ew)     
          let [c1,c2,c3,c4] = take 4 $ drop (2*(i-1)) controls
          setXPos c1 (a+  (p-a)/3)
          setXPos c2 (a+2*(p-a)/3)
          setXPos c3 (p+  (b-p)/3)
          setXPos c4 (p+2*(b-p)/3)
{-
        let (a,b) = ( round (p * fromIntegral w) - 6 , round (q * fromIntegral h) - 6 )     
        -- wxwidgets measures mouseleftdrag relative to the windows position,
        -- even if the windows is moved meantime....      
        set win [ position := Point a b ]
        swapMVar pos (p,q)
-}
       

        return s
      MouseLeftUp (Point x y) mod -> do 
        windowReleaseMouse win
        return Up 
      _ -> return s 
  putMVar state s'         
  --repaint parent 
  
-------------------------------------------------------
      
-- | Envelope widget. Instances: 'Widget', 'Dimensions', 'Identity', 'Tipped'. 
data EnvelopeWidget = EnvelopeWidget
  { _container   :: Panel ()
  , _window      :: Panel ()
  , _maxsegments :: Int
  , _signed      :: Bool
  , _envelope    :: MVar (Envelope Double Double)
--  , _state    :: MVar EnvelopeState
  , _endpoints   :: MVar [ControlPoint]
  , _controls    :: MVar [ControlPoint]
  }

{-  
data EnvState 
  = Up 
  | Down { _startPos :: (Int,Int) , _startValue :: Int }
-}

-------------------------------------------------------
 
{-   
instance Widget EnvelopeWidget where
  widget env = widget (_window env)    
    
instance Dimensions EnvelopeWidget where
  outerSize   = mapAttrW _window outerSize
  position    = mapAttrW _window position
  area        = mapAttrW _window area
  bestSize    = mapAttrW _window bestSize  
  clientSize  = mapAttrW _window clientSize
  virtualSize = mapAttrW _window virtualSize

instance Tipped EnvelopeWidget where
  tooltip = mapAttrW _window tooltip

instance Identity EnvelopeWidget where
  identity = mapAttrW _window identity
-}

instance Widget EnvelopeWidget where
  widget env = widget (_container env)    
    
instance Dimensions EnvelopeWidget where
  outerSize   = mapAttrW _container outerSize
  position    = mapAttrW _container position
  area        = mapAttrW _container area
  bestSize    = mapAttrW _container bestSize  
  clientSize  = mapAttrW _container clientSize
  virtualSize = mapAttrW _container virtualSize

instance Tipped EnvelopeWidget where
  tooltip = mapAttrW _container tooltip

instance Identity EnvelopeWidget where
  identity = mapAttrW _container identity
  
-------------------------------------------------------
  
makeEnvelopeWidget 
  :: Window a               -- ^ parent window
  -> Int                    -- ^ maximal number of segments (8 is a sane default)
  -> Bool                   -- ^ signed envelope? (that is, range is [-1,1] or [0,1]) 
  -> [Prop EnvelopeWidget]  -- ^ properties (eg. 'selection' or 'tooltip') 
  -> IO EnvelopeWidget 

makeEnvelopeWidget parent maxsegments signed props = do 
  container <- panel parent []

  small <- panel container []
  
  spin <- spinCtrl small 1 maxsegments [ text := "1" , tooltip := "# of segments" ]
  status <- staticText small [ text := "try right click, too" ]
  
  win <- panelCreate container 0 (Rect 0 0 50 50) 0 
      
  let dseg = if signed then defaultSegment' 0 else defaultSegment' (0.5)
      --dseg = EnvSegment LinearS 0.5 0.5 0.5 1
      --envelope = foldr (uncurry ECons) (ESeg dseg) [ (dseg,fromIntegral i) | i<-[1..maxsegments-1] ] 
      envelope' = [ (dseg, fromIntegral i) | i<-[1..maxsegments] ]
      envelope = Envelope 1 envelope'
  menv <- newMVar envelope

  -- these MVar-s are necessary only because of the 
  -- impossibility of "tying the knot" here
  eps <- newMVar []
  cts <- newMVar []

  let env = EnvelopeWidget container win maxsegments signed menv eps cts 
  
  let f i = do
        cp <- makeControlPoint env Endpoint signed (i,0) (i<=1) []
        setPos cp (nthEndpoint i envelope')
        return cp
      g i j = do
        cp <- makeControlPoint env Control  signed (i,j) False [] -- (i==0) [] -- False  []
        setPos cp (nthControl i j envelope')
        return cp
  endpoints <- sequence [ f i   | i <- [0..maxsegments  ] ]
  controls  <- sequence [ g i j | i <- [0..maxsegments-1], j<-[1,2] ]

  swapMVar (_endpoints env) endpoints 
  swapMVar (_controls  env) controls 

  set win 
    [ on paint  := paintEnvelope env 
 --   , on mouse := mouseEnvelope win minmax 
    , on clickRight := changeType env
    , on resize := resizeEnvelope env
    , visible   := True
    ]
  set env props

  set spin [ on select := spinSelect env spin maxsegments menv win endpoints controls ]

  set small 
    [ layout := hfill $ row 0
      [ floatLeft (widget status)
      , hglue
      , widget spin
      ]
    ]

  set container 
    [ layout := column 5 
      [ fill (widget win) 
      , widget small
      ]
    , on resize := resizeContainer container win small 
    ]

{-  
-- here i censored myself by deleting four lines of extreme vulgarity about 
-- just how much the totally non-working layout "engine" of wxwidgets sucks.

  set container 
    [ layout := column 5 
      [ fill (widget win) 
      , hrule 100
      , hfill $ row 0
        [ floatLeft (widget status)
        , hglue
        , rigid (widget spin)
        ]
      ]
    ]
-}

  return $ env

-------------------------------------------------------

changeType ew (Point x y) = 
  do
    Rect l t w h <- windowGetRect (_window ew)
    let a = fromIntegral x / fromIntegral w 
    Envelope n envl <- takeMVar (_envelope ew)
    let (envl',i) = worker a envl 0
    putMVar (_envelope ew) $ Envelope n envl'
    controls <- readMVar (_controls ew)
    forM_ controls $ \cp -> when (fst (cp_index cp) == i) $ set cp [ enabled :~ not ]
    repaint (_window ew)
    repaintAllControls ew
  where
    worker a (this@(es,t):rest) i = if a<=t 
      then ( (flipType es,t):rest , i ) 
      else let (xs,i') = worker a rest (i+1) in (this:xs,i')   
    worker _ [] i = ([],i)       
       
-------------------------------------------------------

double [] = []
double (x:xs) = x:x:double xs

spinSelect ew spin maxsegments menv win endpoints controls = do
  -- the clamping is unneccessary in theory, but i don't trust wx
  n <- liftM (clamp (1,maxsegments)) $ get spin selection
  adjustMVar menv $ updateNumberOfSegments n 

  readMVar menv >>= \env@(Envelope _ envl) -> do
    let f cp = let (i,_) = cp_index cp in setPos cp (nthEndpoint i   envl)
        g cp = let (i,j) = cp_index cp in setPos cp (nthControl  i j envl)
    mapM_ f endpoints 
    mapM_ g controls   
    forM_ (zip  [0..] endpoints    ) $ \(i,cp   ) -> set cp [ enabled := (i<=n) ]
    forM_ (zip3 [1..] controls (double envl)) $ \(i,cp,es) -> set cp [ enabled := (i<=2*n) && isCubic (fst es) ]
    repaintAllControls ew
    
  repaint win
  
resizeContainer container win small = do
  Size w h <- get container clientSize
  Size _ y <- get small     outerSize
  set win   [ position := Point 0 0     , outerSize := Size w (h-y-5) ]
  set small [ position := Point 0 (h-y) , outerSize := Size w y       ]
  return ()
  
-------------------------------------------------------

pairs :: [a] -> [(a,a)]
pairs (x:xs@(y:_)) = (x,y) : pairs xs
pairs [] = []
pairs [x] = []

-- col2 is the color of the tangent lines
drawBezier :: Envelope Double Double -> Bool -> Color -> Color -> Int -> DC a -> Rect -> IO ()
drawBezier env signed col col2 penwidth dc rect@(Rect l t w h) = 
  do
    dcDraw dc $ do 
      -- draw tangents
      dcWithPenStyle dc (penColored col2 penwidth) $ do
        let Envelope n envl' = env
            envl = take n envl'
        forM_ (zip envl [1..n]) $ \(es,k) -> when (isCubic $ fst es) $ do
          dcDrawLine dc (pt $ nthEndpoint (k-1) envl') (pt $ nthControl (k-1) 1 envl')  
          dcDrawLine dc (pt $ nthEndpoint (k  ) envl') (pt $ nthControl (k-1) 2 envl')  
      -- draw the curve
      dcWithPenStyle dc (penColored col  penwidth) $ do 
        forM_ (pairs points) $ \(x,y) -> dcDrawLine dc x y 
  where  
    pt :: (Double,Double) -> Point
    pt (x,y) = Point (round $ x * fromIntegral w) (h - f y) 
    f :: Double -> Int
    f x = if signed 
      then round $ 0.5*(1-x)*fromIntegral h 
      else round $     (1-x)*fromIntegral h 
    g :: Int -> Double 
    g i = calcEnvelope ( (0.5 + fromIntegral i) / fromIntegral w) env 
    points = [ Point i (h-(f (g i))) | i<-[0,4..w] ] 

drawBezier1 env col col2 penwidth = do
  let win = _window env
  rect <- windowGetRect win
  withClientDC win $ \dc -> do
    
    Control.Monad.when (_signed env) $ do
      Size w h <- windowGetClientSize (_window env)
      let hh = h `div` 2
      dcWithPenStyle dc (penColored (colorRGB 100 100 200) 1) $ do
        dcDrawLine dc (Point 0 hh) (Point w hh)

    e <- readMVar (_envelope env) 
    drawBezier e (_signed env) col col2 penwidth dc rect
   
-------------------------------------------------------
   
repaintAllControls env = do
  eps <- readMVar (_endpoints env)
  cts <- readMVar (_controls  env)
  forM_ (eps++cts) $ \cp -> readMVar (cp_enabled cp) >>= \b -> when b $ repaint (cp_win cp)   
   
backgroundColor = (colorRGB 60  60  100)  
bezierColor     = (colorRGB 255 200 200)
tangentColor    = (colorRGB 50  150 100)
bezierWidth = 1
    
paintEnvelope :: EnvelopeWidget -> DC () -> Rect -> IO ()   
paintEnvelope ew dc rect = do
  let bs = brushSolid backgroundColor
  dcWithBrushStyle dc bs $ dcDrawRectangle dc rect
  drawBezier1 ew bezierColor tangentColor bezierWidth
     
-------------------------------------------------------

resizeEnvelope env = do
  eps <- readMVar (_endpoints env)
  cts <- readMVar (_controls  env)
  forM_ (eps++cts) updatePos

-------------------------------------------------------
