
-- | Timeline widget for wxHaskell. Useful for animation tools.
-- TODO: some serious cleaning & refactoring. 
-- The code is ugly as hell. But it seems to work.

{-# LANGUAGE ScopedTypeVariables, MultiParamTypeClasses, FlexibleInstances, FunctionalDependencies #-}

module Graphics.UI.WX.Controls.TimelineWidget
  ( 
    TimelineWidget
  , makeTimelineWidget 
{-  
  , nrows
  , duration
  , zoom
  , blocks
-}

  , module Data.Timeline
  
  , OnZoomScrollFunc
  , OnSelectFunc
  , OnMoveFunc
  , OnResizeFunc
  , OnInsertFunc

{-  
  , onZoomScroll
  , onBlockSelect
  , onBlockMove
  , onBlockResize
  , onBlockInsert
-}

  , TimelineDisplay(..)
  
  ) where

import Control.Monad
import Control.Concurrent.MVar

import Data.Array.IArray
import Data.Bits
import Data.Maybe
import Data.List hiding (intersect)

{-
import Data.Map (Map)
import qualified Data.Map as Map
-}

import Graphics.UI.WX 
import Graphics.UI.WXCore hiding (Time)

import Data.Timeline

-------------------------------------------------------

clamp :: (Int,Int) -> Int -> Int
clamp (minv,maxv) x = max (min x maxv) minv

-------------------------------------------------------

type FeedbackFunc     = String -> IO ()
type OnSelectFunc a t = (Int, Block a t) -> IO ()
type OnMoveFunc   a t = (Int, Block a t) -> (Int, Iv t) -> IO ()
type OnResizeFunc a t = (Int, Block a t) -> Iv t -> IO ()
type OnInsertFunc a t = Int -> Iv t -> IO (Maybe (Block a t))
type OnInsertSuccessFunc a t = (Int, Block a t) -> IO ()
type OnZoomScrollFunc t = Double -> (t,t) -> Size -> IO ()

type Time = Int
  
-- | TimelineWidget of blocks of type a
data TimelineWidget a = TimelineWidget 
  { _window    :: ScrolledWindow ()
  , _tline     :: MVar (Timeline a Time) 
  , _rowheight :: Int
  , _zoom      :: MVar Double        -- 1.0 = 100%; 2.0 = 50%; etc
  , _state     :: MVar (MouseState a)
  , _dragwin   :: Panel ()

  , _feedback  :: MVar FeedbackFunc  
  , _onselect  :: MVar (OnSelectFunc a Time)
  , _onmove    :: MVar (OnMoveFunc   a Time)
  , _onresize  :: MVar (OnResizeFunc a Time)
  , _oninsert  :: MVar (OnInsertFunc a Time)
  , _oninsertsuccess  :: MVar (OnInsertSuccessFunc a Time)
  , _onzoomscroll :: MVar (OnZoomScrollFunc Time)
  
  , _filtermemo :: MVar ( Rect , Filtered a )  -- memoization for speed
  }
  
feedb :: TimelineWidget a -> String -> IO ()
feedb tlw s = readMVar (_feedback tlw) >>= \callback -> callback s 
  
-------------------------------------------------------
  
makeAttributeFromMVar :: String -> (TimelineWidget b -> MVar a) -> Attr (TimelineWidget b) a
makeAttributeFromMVar name f = 
  newAttr name getter setter where
    getter tlw   = readMVar (f tlw)
    setter tlw x = do
      swapMVar (f tlw) x
      return ()
      --resizeTimelineWidget tlwidget
      

class TimelineDisplay widg a | widg -> a where
  -- | The default feedback mechanism is "logStatus". 
  feedback      :: Attr widg FeedbackFunc           
  -- | Selection callback
  onBlockSelect :: Attr widg (OnSelectFunc a Time)  
  -- | Move callback
  onBlockMove   :: Attr widg (OnMoveFunc   a Time)  
  -- | Resize callback
  onBlockResize :: Attr widg (OnResizeFunc a Time)  
  -- | Insert callback (default is @\_ _ -> return Nothing@, thus no insert)                                      
  onBlockInsert :: Attr widg (OnInsertFunc a Time)  
  -- | Callback on successful insert
  onBlockInsertSuccess :: Attr widg (OnInsertSuccessFunc a Time) 
  -- | Zoom \/ scroll callback
  onZoomScroll :: Attr widg (OnZoomScrollFunc Time)  
  -- | Number of rows (levels). Read-only. Note: the numbering starts from 1.
  nrows    :: Attr widg Int  
  -- | Size of the timeline. Read-only.
  duration :: Attr widg Int  
  -- | The blocks (events) themselves.  
  blocks   :: Attr widg (Timeline a Time)  
  -- | Zoom level. 1.0 = 100%, 2.0 = 50%, etc.
  zoom     :: Attr widg Double 
  

instance TimelineDisplay (TimelineWidget a) a where

  --  The default feedback mechanism is "logStatus". 
  --feedback :: Attr (TimelineWidget a) FeedbackFunc
  feedback = makeAttributeFromMVar "feedback" _feedback
  
  --  Selection callback
  --onBlockSelect :: Attr (TimelineWidget a) (OnSelectFunc a Time)
  onBlockSelect = makeAttributeFromMVar "onBlockSelect" _onselect
      
  --  Move callback
  --onBlockMove   :: Attr (TimelineWidget a) (OnMoveFunc   a Time)
  onBlockMove   = makeAttributeFromMVar "onBlockMove"   _onmove
  
  --  Resize callback
  --onBlockResize :: Attr (TimelineWidget a) (OnResizeFunc a Time)
  onBlockResize = makeAttributeFromMVar "onBlockResize" _onresize
  
  --  Insert callback (default is @\_ _ -> return Nothing@, thus no insert)
  --onBlockInsert :: Attr (TimelineWidget a) (OnInsertFunc a Time)
  onBlockInsert = makeAttributeFromMVar "onBlockInsert" _oninsert
  
  --  Callback on successful insert
  --onBlockInsertSuccess :: Attr (TimelineWidget a) (OnInsertSuccessFunc a Time)
  onBlockInsertSuccess = makeAttributeFromMVar "onBlockInsertSuccess" _oninsertsuccess
  
  --  Zoom \/ scroll callback
  --onZoomScroll :: Attr (TimelineWidget a) (OnZoomScrollFunc Time)
  onZoomScroll = makeAttributeFromMVar "onZoomScroll" _onzoomscroll
    
  --  Number of rows (levels). Read-only. Note: the numbering starts from 1.
  --nrows :: Attr (TimelineWidget b) Int
  nrows = newAttr "nrows" getter setter where
    getter tlwidget   = liftM tlNumberOfRows $ readMVar (_tline tlwidget)
    setter tlwidget x = do
      --return ()
      error "nrows/set"
    {-
      swapMVar (_nrows tlwidget) x
      resizeTimelineWidget tlwidget
    -}
    
  --  Size of the timeline. Read-only.
  --duration :: Attr (TimelineWidget b) Int
  duration = newAttr "duration" getter setter where
    getter tlwidget   = liftM tlDuration $ readMVar (_tline tlwidget)
    setter tlwidget x = do
      --return ()
      error "durations/set"
  {-
      swapMVar (_duration tlwidget) x
      resizeTimelineWidget tlwidget
  -}
  
  
  --  The blocks (events) themselves.
  --blocks :: Attr (TimelineWidget b) (Timeline b Time)
  blocks = newAttr "blocks" getter setter where
    getter tlwidget   = readMVar (_tline tlwidget)
    setter tlwidget x = do
      swapMVar (_tline tlwidget) x 
      invalidateFilterMemo tlwidget -- !!
      repaint (_window tlwidget)
  
-------------------------------------------------------

  --  Zoom level. 1.0 = 100%, 2.0 = 50%, etc.
  --zoom :: Attr (TimelineWidget b) Double
  zoom = newAttr "zoom" getter setter where
    getter tlwidget   = readMVar (_zoom tlwidget)
    setter tlwidget x = do 
  
      let win = _window tlwidget
      oz  <- readMVar (_zoom tlwidget) 
      dur <- get tlwidget duration -- readMVar (_duration tlwidget)
      Point ox0 y0 <- scrolledWindowGetViewStart win
      Size  oxs ys <- scrolledWindowGetVirtualSize win
      Size w _ <- windowGetClientSize (_window tlwidget)
  
      let crit = fromIntegral dur / fromIntegral w 
          nz = if crit > 1   
            then min crit (max 1 x)
            else crit -- 1.0 / crit  
  
      let nxs = (round $ fromIntegral dur / nz)
      let nx0 = if crit > 1 
            then let stuff = round $ 
                       ( fromIntegral ox0 + 0.5 * fromIntegral w ) * oz / nz
                       - 0.5 * fromIntegral w 
                     mx = floor $ fromIntegral dur - fromIntegral w / nz 
                 in  clamp (0,mx) stuff 
            else 0
  
      --logStatus $ show (crit,oz,nz,(ox0,y0),(oxs,ys),nx0)
       
      invalidateFilterMemo tlwidget -- !!
      Control.Monad.when (nz /= oz) $ do
        swapMVar (_zoom tlwidget) nz 
        scrolledWindowSetScrollbars win 1 1 nxs ys nx0 y0 False       
        -- resizeTimelineWidget' tlwidget
        repaint win
  
      let (_,s) = fromScreenXY nz 1 (Point  nx0    0)
          (_,t) = fromScreenXY nz 1 (Point (nx0+w) 0)
      siz <- windowGetClientSize win
      readMVar (_onzoomscroll tlwidget) >>= \callback -> callback nz (s,t) siz  -- ??
  
-------------------------------------------------------


scrollTimelineWidget tlwidget oldhandler evt = do
  oldhandler evt
  -- ok, the old handler is empty, and the real scroll handler will run after us
  -- so again, we need to hack. why am i not surprised...
  case evt of
    ScrollTrack Horizontal newxpos -> do
      let win = _window tlwidget
      Point _ y <- scrolledWindowGetViewStart win
      Size xs ys <- scrolledWindowGetVirtualSize win
      scrolledWindowSetScrollbars win 1 1 xs ys newxpos y False  -- True is nofresh (but does not work...)
      set tlwidget [ zoom :~ id ]   -- ez meghivja a ZoomScroll callbacket
    _ -> return ()
{-    
  pt <- scrolledWindowGetViewStart (_window tlwidget)
  logStatus (show evt ++ " | " ++ show pt)
-}

-------------------------------------------------------

setCursorResize :: Window a -> IO ()
setCursorResize win = do            
  cursor <- cursorCreateFromStock wxCURSOR_SIZEWE
  windowSetCursor win cursor   
  return ()

setCursorNormal :: Window a -> IO ()
setCursorNormal win = do            
  cursor <- cursorCreateFromStock wxCURSOR_ARROW
  windowSetCursor win cursor   
  return ()

-------------------------------------------------------

-- x1 y1 x2 y2, not x1 y1 width height
data Rect1 = Rect1 !Int !Int !Int !Int deriving Show

type Filtered b = [ ( Rect1 , Rect1 , Int , Block b Time ) ]

toRect1    (Rect x0 y0 w  h ) = Rect1 x0 y0 (x0+w ) (y0+h )
fromRect1 (Rect1 x0 y0 x1 y1) = Rect  x0 y0 (x1-x0) (y1-y0)

inRect1 :: Point -> Rect1 -> Bool
(Point x y) `inRect1` (Rect1 x0 y0 x1 y1) 
  =  (x0 <= x)
  && (x <= x1)
  && (y0 <= y)
  && (y <= y1)

invalidateFilterMemo :: TimelineWidget b -> IO ()
invalidateFilterMemo tlw = swapMVar (_filtermemo tlw) (Rect 0 0 0 0, [] ) >> return ()

calcFilterScreenRect :: TimelineWidget b -> Rect -> IO () -- [(Rect1,Rect1,Int,Block b Time)]
calcFilterScreenRect tlwidget rect = do

  zoom  <- readMVar (_zoom tlwidget)
  let rh = _rowheight tlwidget
      screen = -- toRect1 rect
               let (Rect1 x1 y1 x2 y2) = toRect1 rect in Rect1 (x1-1) y1 x2 y2
               -- bal szelso doboz szele render hekk ^^^^^^
                
  -- nrows <- readMVar (_nrows tlwidget)
  -- dur   <- readMVar (_duration tlwidget)
  tline <- readMVar (_tline tlwidget)
  let arr = _tlarr tline
      bnds@(nrowsmin,nrowsmax) = bounds arr
      
  let f lev (SL blocks') = mapMaybe g blocks where
        blocks = map (\bl -> toScreenRect zoom rh (lev,bl)) blocks'
        g (rect1,lev,b) = do { r <- intersectRect1 screen rect1 ; return (rect1,r,lev,b) } 

  let result = concat [ f lev (arr!lev) | lev<-[nrowsmin..nrowsmax] ]
  swapMVar (_filtermemo tlwidget) (rect,result)
  return ()

filterScreenRect :: TimelineWidget b -> IO (Filtered b)
filterScreenRect tlwidget = do
  rect@(Rect x0 y0 w h) <- windowGetViewRect (_window tlwidget)  
  (oldrect,memo) <- readMVar (_filtermemo tlwidget)
  if rect == oldrect 
    then return memo
    else do 
      calcFilterScreenRect tlwidget rect
      liftM snd $ readMVar (_filtermemo tlwidget)
      
{-
filterScreenRect :: TimelineWidget b -> Rect -> IO Filtered -- [(Rect1,Rect1,Int,Block b Time)]
filterScreenRect tlwidget rect = do
  zoom  <- readMVar (_zoom tlwidget)
  let rh = _rowheight tlwidget
      screen = -- toRect1 rect
               let (Rect1 x1 y1 x2 y2) = toRect1 rect in Rect1 (x1-1) y1 x2 y2
               -- bal szelso doboz szele render hekk ^^^^^^
                
  -- nrows <- readMVar (_nrows tlwidget)
  -- dur   <- readMVar (_duration tlwidget)
  tline <- readMVar (_tline tlwidget)
  let arr = _tlarr tline
      bnds@(nrowsmin,nrowsmax) = bounds arr
      
  let f lev (SL blocks') = mapMaybe g blocks where
        blocks = map (\bl -> toScreenRect zoom rh (lev,bl)) blocks'
        g (rect1,lev,b) = do { r <- intersectRect1 screen rect1 ; return (rect1,r,lev,b) } 

  return $ concat [ f lev (arr!lev) | lev<-[nrowsmin..nrowsmax] ]
-}


fromScreenXY :: Double -> Int -> Point -> ( Int , Time )
fromScreenXY zoom rh (Point x y) = ( lev , time ) where
  minlevel = 1
  lev = minlevel + div (y-1) rh 
  time = round $ fromIntegral x * zoom

toScreenPos :: Double -> Int -> Int -> Int -> (Int,Int)
toScreenPos zoom rh level t = (x,y) where 
  minlevel = 1
  y = rh * (level - minlevel) + 1  
  x = round $ fromIntegral t / zoom

toScreenRect :: Double -> Int -> (Int , Block b Time) -> (Rect1, Int, Block b Time) 
toScreenRect zoom rh (level,block) = ( rect , level , block ) where
  minlevel = 1
  rect = Rect1 x1 y1 x2 y2 
  y1 = rh * (level - minlevel) + 1  
  y2 = y1 + rh - 1
  (s,t) = _block_iv block
  x1 = floor   $ fromIntegral s / zoom
  x2 = ceiling $ fromIntegral t / zoom
    
intersectIv :: (Int,Int) -> (Int,Int) -> Maybe (Int,Int)
intersectIv (a,b) (c,d) = 
  if f>=e 
    then Just (e,f) 
    else Nothing 
  where
    e = max a c    
    f = min b d
    
intersectRect1 :: Rect1 -> Rect1 -> Maybe Rect1 
intersectRect1 (Rect1 x1 y1 x2 y2) (Rect1 u1 v1 u2 v2) = do
  (p1,p2) <- intersectIv (x1,x2) (u1,u2)
  (q1,q2) <- intersectIv (y1,y2) (v1,v2)
  return $ Rect1 p1 q1 p2 q2

insideIv x (a,b) = x >= a && x <= b  

insideRect1 (x,y) (Rect1 p1 q1 p2 q2) = insideIv x (p1,q1) && insideIv y (p2,q2)

-------------------------------------------------------
    
instance Widget (TimelineWidget b) where
  widget tlwidget = widget (_window tlwidget)    
    
instance Dimensions (TimelineWidget b) where
  outerSize   = mapAttrW _window outerSize
  position    = mapAttrW _window position
  area        = mapAttrW _window area
  bestSize    = mapAttrW _window bestSize  
  clientSize  = mapAttrW _window clientSize
  virtualSize = mapAttrW _window virtualSize

instance Sized (TimelineWidget b) where
  size = mapAttrW _window size

instance Tipped (TimelineWidget b) where
  tooltip = mapAttrW _window tooltip

instance Identity (TimelineWidget b) where
  identity = mapAttrW _window identity
  
-------------------------------------------------------
  
makeTimelineWidget 
  :: forall x a. 
  Window x -- ^ parent 
  -> Int      -- ^ number of rows 
  -> (Time,Time)  -- ^ start/end
  -> Int      -- ^ height of a row in pixels
  -> [Prop (TimelineWidget a)] 
  -> IO (TimelineWidget a) 
  
makeTimelineWidget parent nrows (start,end) rowheight props = do 
  let header = 20
      rect   = Rect 0 0 512 200 --duration (nrows*rowheight)
      zoom   = 1.0 :: Double  
  win <- scrolledWindowCreate parent 0 rect 0
  
  dragwin <- panel win [ size := Size 100 (rowheight+1) , visible := False ]
  set dragwin [ on paint := dragPaint dragwin ]
  -- windows hacking
  flg <- windowGetWindowStyleFlag dragwin
  windowSetWindowStyleFlag dragwin $ 
    flg .&. (complement wxNO_FULL_REPAINT_ON_RESIZE) 
    -- .|. wxTRANSPARENT_WINDOW
 
  feedback <- newMVar logStatus
  onselect <- newMVar $ \_   -> return ()
  onmove   <- newMVar $ \_ _ -> return ()
  onresize <- newMVar $ \_ _ -> return ()
  oninsert <- newMVar $ \_ _ -> return Nothing
  oninsertsucc <- newMVar $ \_ -> return ()
  onzoomsc <- newMVar $ \_ _ _ -> return ()
  
  mstate  <- newMVar Up
  mblocks <- newMVar $ emptyTimeline (start,end) (1,nrows) :: IO (MVar (Timeline a Time))
  mzoom   <- newMVar zoom
  
  memo <- newMVar (Rect 0 0 0 0, []) :: IO (MVar (Rect , Filtered a))
  
  let tlwidget = TimelineWidget 
                   win mblocks rowheight mzoom mstate dragwin
                   feedback onselect onmove onresize oninsert oninsertsucc onzoomsc
                   memo
  resizeTimelineWidget tlwidget
  set win 
    [ on paint  := paintTimelineWidget  tlwidget  
    , on resize := resizeTimelineWidget tlwidget
    , on mouse  := mouseTimelineWidget  tlwidget 
    , visible   := True
    ]

  oldhandler <- windowGetOnScroll win
  windowOnScroll win (scrollTimelineWidget tlwidget oldhandler)
  
{-      
  -- capturing the mouse solves the problem
  set dragwin 
    [ on mouse := \evt -> case evt of
        MouseLeave {} -> return ()
        MouseEnter {} -> return ()
        _ -> do
          -- wxWidgets developers cannot code proper event propagation
          evtHandler <- get win (on mouse)
          let pt@(Point xa ya) = mousePos evt
          Point x0 y0 <- get dragwin position
          let unscrolled = Point (xa+x0) (ya+y0)
          scrolled <- scrolledWindowCalcUnscrolledPosition win unscrolled
          evtHandler (replaceEventMousePos evt scrolled)
          -- propagateEvent    
    ]
-}

  set tlwidget props
  return tlwidget

-------------------------------------------------------

{-
-- now this one is wxHaskell's problems and not wxWidgets' one
replaceEventMousePos :: EventMouse -> Point -> EventMouse
replaceEventMousePos evt new = case evt of
  MouseMotion      point modif   -> MouseMotion new modif
  MouseEnter       point modif   -> MouseEnter       new modif 
  MouseLeave       point modif   -> MouseLeave       new modif 
  MouseLeftDown    point modif   -> MouseLeftDown    new modif 
  MouseLeftUp      point modif   -> MouseLeftUp      new modif 
  MouseLeftDClick  point modif   -> MouseLeftDClick  new modif 
  MouseLeftDrag    point modif   -> MouseLeftDrag    new modif 
  MouseRightDown   point modif   -> MouseRightDown   new modif 
  MouseRightUp     point modif   -> MouseRightUp     new modif 
  MouseRightDClick point modif   -> MouseRightDClick new modif 
  MouseRightDrag   point modif   -> MouseRightDrag   new modif 
  MouseMiddleDown  point modif   -> MouseMiddleDown  new modif 
  MouseMiddleUp    point modif   -> MouseMiddleUp    new modif 
  MouseMiddleDClick point modif  -> MouseMiddleDClick new modif
  MouseMiddleDrag  point modif   -> MouseMiddleDrag  new modif 
  MouseWheel     b point modif   -> MouseWheel     b new modif
-}

-------------------------------------------------------

resizeTimelineWidget tlwidget = do
--  set tlwidget [ zoom :~ id ]   -- normalizalunk vagy mi
  resizeTimelineWidget' tlwidget  
  set tlwidget [ zoom :~ id ]   -- normalizalunk vagy mi
  repaint (_window tlwidget)
    
-- set zoom meg ezt hivja, azert..
resizeTimelineWidget' tlwidget = do
  let win = _window tlwidget
      rowheight = _rowheight tlwidget
  duration <- get tlwidget duration 
  nrows <- liftM tlNumberOfRows $ readMVar (_tline tlwidget)
  z <- get tlwidget zoom  
  --zoom <- readMVar (_zoom tlwidget)
  
  --scrolledWindowAdjustScrollbars win
  Point x y <- scrolledWindowGetViewStart win
  scrolledWindowSetScrollbars win 1 1 (round $ fromIntegral duration / z) (nrows*rowheight+1) x y False
  --repaint win

-------------------------------------------------------

dragPaint win dc rect = do
  let brush = BrushStyle (BrushHatch HatchCrossDiag) (colorRGB 100 100 100)
  dcWithBrushStyle dc brush $ do
    dcDrawRectangle dc rect
      
-------------------------------------------------------

sizeLimit :: Double -> Int
sizeLimit z = min 16 (max 5 (ceiling $ 20.0 / z))
  
data MouseState a
  = Up 
  | Drag
      { _startPos :: (Int,Int) 
      , _offset   :: (Int,Int)
      , _origlevel :: Int
--      , _newPos   :: (Int,Int) 
      , _theblock :: Block a Time 
      }
  | Resize
      { _startPos  :: (Int,Int) 
      , _offset    :: (Int,Int)
      , _origlevel :: Int
      , _theblock  :: Block a Time
      , _leftside  :: Bool       -- left or right side
      , _origscreen :: (Int,Int) -- original interval on the screen 
      }
  | Insert
      { _startPos  :: (Int,Int) 
      , _origlevel :: Int
      , _startx    :: Int
      }         

mouseTimelineWidget tlwidget event = do
  let win     = _window tlwidget
      dragwin = _dragwin tlwidget
      rh      = _rowheight tlwidget

  winrect@(Rect wx0 wy0 _ _) <- windowGetViewRect win
  
  s <- takeMVar (_state tlwidget)
  s' <- case s of
    
    Up -> case event of
    
      MouseWheel down pt modif -> do
        let a = 0.15
            ch = exp (if down then a else (-a))
        invalidateFilterMemo tlwidget  -- !!!!
        set tlwidget [ zoom :~ (*ch) ]
        repaint win
        return s
            
      MouseLeftDown pt@(Point xp yp) modif -> do
        windowSetFocus win -- wxWidgets on Windows sucks ass
        filtblocks <- filterScreenRect tlwidget -- winrect 
        newstate <- case find (\(_,rect,_,_) -> pt `inRect1` rect) filtblocks of
        
          Nothing -> do
            --logStatus $ "[" ++ show pt ++ "]"
            z <- get tlwidget zoom
            let (tlev,tx) = fromScreenXY z rh $ Point xp yp
            
            maxlev <- get tlwidget nrows 
            let minlev = 1
            if tlev >= minlev && tlev <= maxlev
              then do
                let (_,y0) = toScreenPos z rh tlev tx
                pos <- scrolledWindowCalcScrolledPosition win (Point xp (y0-1))
                set dragwin [ position := pos , size := Size 0 (rh+1) ] 
                windowCaptureMouse win
                return $ Insert -- s
                  { _startPos = (xp,yp)
                  , _origlevel = tlev
                  , _startx = tx
                  }
              else
                return Up
              
          Just ( rect@(Rect1 x0 y0 x1 y1) , _ , oldlev , block ) -> do
            --logStatus $ show rect ++ " --- " ++ show pt ++ " --- " ++ show (_block_name block)
            pos <- scrolledWindowCalcScrolledPosition win (Point x0 (y0-1))
            let siz = Size (x1-x0+1) (y1-y0+2)
            set dragwin [ {- visible := True ,-} position := pos , size := siz ]
  
            windowCaptureMouse win
  
            let drag0   = Drag   undefined undefined undefined undefined
                resize0 = Resize undefined undefined undefined undefined undefined undefined
                f x0 x1 xp 
                  | x1 - x0 < 15   = drag0
                  | xp - x0 < 5    = resize0 { _leftside = True  , _origscreen = (x0,x1) }
                  | x1 - xp < 5    = resize0 { _leftside = False , _origscreen = (x0,x1) }
                  | otherwise      = drag0
                state0 = f x0 x1 xp
                
            case state0 of 
              Resize {} -> setCursorResize win
              _         -> return ()
      
            readMVar (_onselect tlwidget) >>= \callback -> callback (oldlev,block)
            return $ state0
              { _startPos = (xp,yp) 
              , _offset   = (xp-x0,yp-y0)
              , _origlevel = oldlev
  --             , _newPos   = (xp,yp) 
              , _theblock = block
              }
        return newstate
        
      _ -> return s 

------ drag ------
      
    down@(Drag{}) -> case event of

{-
-- a buta wxwidgets,
-- (kisebb baj hogy ha child ablakra ha ra/lejovok akkor leave/enter),
-- MASRESZT NEM VESZI MINDIG ESZRE HA KIMEGYEK AZ ABLAKBOL?!!
      MouseLeave pt _ -> do
        --set (_dragwin tlwidget) [ visible := False ]      
        logStatus $ "leave " ++ show pt
        return s --Up
-}
  
      MouseLeftDrag pt@(Point xp yp) modif -> do
        let (xofs,yofs) = _offset s
            pt = Point (xp-xofs) (yp-yofs)
            
        z <- get tlwidget zoom
        let (tlev,tx) = fromScreenXY z rh $ Point (xp-xofs) (yp-yofs + div rh 2)
        feedb tlwidget $ "target @ " ++ show tx ++ " (level " ++ show tlev ++ ")"
        
        pos <- scrolledWindowCalcScrolledPosition win pt
        set dragwin [ visible := True , position := pos ]  
        let newstate = s -- { _newPos = (xp,yp) } 
        return newstate

      MouseLeftUp pt@(Point xp yp) modif -> do
        zoom <- readMVar (_zoom tlwidget)
        tline <- readMVar (_tline tlwidget)
        let (xofs,yofs) = _offset s
            rh = _rowheight tlwidget
            (levmin,levmax) = bounds (_tlarr tline)
            (lev,time0) = fromScreenXY zoom rh $ Point ( xp - xofs ) (yp - yofs + div rh 2 )
            block = _theblock s
            time1 = time0 + _block_duration block
        Control.Monad.when ( (xp,yp) /= (_startPos s) ) $ 
          if (lev >= levmin) && (lev <= levmax)
            then do
              let oldlev = _origlevel s
                  newiv = (time0,time1)
                  ei_tline' = moveBlock Move (oldlev,_block_iv block) (lev,newiv) Nothing tline                
              case ei_tline' of
                Left err -> feedb tlwidget err
                Right tl -> do
                  set tlwidget [ blocks := tl ]
                  invalidateFilterMemo tlwidget -- !!!
                  feedb tlwidget $ "drag-and-drop ok" -- ++ show (yp-yofs) ++ " " ++ show lev
                  readMVar (_onmove tlwidget) >>= \callback -> callback (oldlev,block) (lev,newiv)
            else do
              feedb tlwidget "level out of range"
              return ()
        set (_dragwin tlwidget) [ visible := False ]      
        windowReleaseMouse win
        return Up
        
      _ -> return s

------ resize ------
        
    down@(Resize{}) -> case event of

      MouseLeftDrag pt@(Point xp yp) modif -> do
      
        let (xs,ys) = _startPos down
            ab0@(a0,b0) = _origscreen down
            w0 = b0 - a0 
            ofs = xp - xs
            (_,yofs) = _offset down 
            
        Point _ y <- get (_dragwin tlwidget) position 
        Size  _ h <- get (_dragwin tlwidget) size 
        z <- get tlwidget zoom

        let 
          ymid = yp -- -yofs + div rh 2
          limit = sizeLimit z 
          ab' = 
            case _leftside down of 
              True  -> ( min (a0 + ofs) (b0-limit) , b0 )            
              False -> ( a0 , max (a0+limit) (b0 + ofs) )
          (tlev,_) = fromScreenXY z rh $ Point a0 ymid 
          ab@(a,b) = 
            ab'
{-            
            if _origlevel down == tlev
              then ab'
              else ab0
-}
          (_,tx1) = fromScreenXY z rh $ Point a ymid 
          (_,tx2) = fromScreenXY z rh $ Point b ymid       
        feedb tlwidget $ "new interval -> " ++ show (tx1,tx2) ++ " | length = " ++ show (tx2-tx1) 
      
        Point x _ <- scrolledWindowCalcScrolledPosition win (Point a y)
        set dragwin [ visible := True , position := Point x y , size := Size (b-a+1) h ]  
        let newstate = s -- { _newPos = (xp,yp) } 
        return newstate
 
      MouseLeftUp pt@(Point xp yp) modif -> do        
        z <- readMVar (_zoom tlwidget)
        tline <- readMVar (_tline tlwidget)

        let (xs,ys) = _startPos down
            ab0@(a0,b0) = _origscreen down
            w0 = b0 - a0 
            ofs = xp - xs
            (_,yofs) = _offset down 
            block = _theblock down
            
        let 
          ymid = yp -- -yofs + div rh 2
          limit = sizeLimit z
          ab@(a,b) = 
            case _leftside down of 
              True  -> ( min (a0 + ofs) (b0-limit) , b0 )            
              False -> ( a0 , max (a0+limit) (b0 + ofs) )
          newiv@(tx1,tx2) = 
            case _leftside down of 
              True  -> ( snd $ fromScreenXY z rh $ Point a ymid , snd (_block_iv block) )
              False -> ( fst (_block_iv block) , snd $ fromScreenXY z rh $ Point b ymid  )
        
        let lev = _origlevel s
            align = if _leftside down then ResizeLeft else ResizeRight
            ei_tline' = moveBlock align (lev,_block_iv block) (lev,newiv) Nothing tline 
                         
        Control.Monad.when ( (xp,yp) /= (_startPos s) ) $        
          case ei_tline' of
            Left err -> feedb tlwidget err
            Right tl -> do
              set tlwidget [ blocks := tl ]
              invalidateFilterMemo tlwidget -- !!!
              feedb tlwidget $ "resize ok "
              readMVar (_onresize tlwidget) >>= \callback -> callback (lev,block) newiv
                  
        set dragwin [ visible := False ]    
        setCursorNormal win                   
        windowReleaseMouse win
        return Up

{-      
        MouseMotion pt@(Point x y) modif -> do
          -- mousemotion gives back logical coordinates already
          case find (\(_,rect,_) -> pt `inRect1` rect) filtblocks of
            Nothing -> feedb tlwidget $ "[" ++ show pt ++ "]"
            Just (_,rect,_) -> feedb tlwidget $ show rect ++ " --- " ++ show pt
          return ()
-}

      _ -> return s

------ insert ------

    down@(Insert {}) -> case event of
    
      MouseLeftDrag pt@(Point xp yp) modif -> do
      
        let (xs,ys) = _startPos down
        z <- get tlwidget zoom

        let 
          a = xs
          limit = sizeLimit z
          b = max (a+limit) xp
          ymid = yp 

          (_,tx1) = fromScreenXY z rh $ Point a ymid 
          (_,tx2) = fromScreenXY z rh $ Point b ymid       
        feedb tlwidget $ "new interval -> " ++ show (tx1,tx2) ++ " | length = " ++ show (tx2-tx1) 
      
        set dragwin [ visible := True , size := Size (b-a+1) (rh+1) ]  
        return s

      MouseLeftUp pt@(Point xp yp) modif -> do
      
        Control.Monad.when ( (xp,yp) /= _startPos down ) $ do
        
          -- copypaste code sucks, but works :)
          
          let (xs,ys) = _startPos down
          z <- get tlwidget zoom
  
          let 
            a = xs
            limit = sizeLimit z
            b = max (a+limit) xp
            ymid = yp 
  
            (_,tx1) = fromScreenXY z rh $ Point a ymid 
            (_,tx2) = fromScreenXY z rh $ Point b ymid       
          
          readMVar (_oninsert tlwidget) >>= \callback -> do
            let lev = _origlevel down 
                iv = (tx1,tx2)
            m <- callback lev iv
            case m of
              Nothing -> return ()
              Just block -> do
                tline <- readMVar (_tline tlwidget)
                let ei = insertBlock ResizeRight lev block tline    
                case ei of
                  Left err -> feedb tlwidget err
                  Right tline' -> do
                    swapMVar (_tline tlwidget) tline'
                    invalidateFilterMemo tlwidget -- !!!
                    readMVar (_oninsertsuccess tlwidget) >>= \callback -> callback (lev,block)
                    return ()
                    
        set dragwin [ visible := False ]    
        windowReleaseMouse win
        return Up

      _ -> return s
      
  putMVar (_state tlwidget) s'
    
    
-------------------------------------------------------

{-
fst3 (x,_,_) = x
snd3 (_,y,_) = y 
thd3 (_,_,z) = z 
-}

-------------------------------------------------------

{-
drawBoxAt dc rect x y = do
  drawLines dc [ Point x y , Point (x+16) y , Point (x+16) (y+16) , Point x (y+16) , Point x y ]
-}

paintTimelineWidget tlwidget dc rect' = do 
  let win = _window    tlwidget
      rh  = _rowheight tlwidget
  rect@(Rect x0 y0 w h) <- windowGetViewRect win
  let a = (y0  ) `div` rh
      b = (y0+h) `div` rh
 
  maxlev <- get tlwidget nrows
  let minlev = 1
      maxy = maxlev*rh

  let background = brushSolid (colorRGB 255 255 240)
  withBrushStyle background $ \brush -> do 
    dcSetBackground dc brush
    let Rect xr yr wr hr = rect'
    dcClearRect dc (Rect xr yr wr (min hr (maxy-yr))) 
 
  forM_ [a..min b maxlev] $ \k -> drawLines dc [ Point x0 (k*rh) , Point (x0+w) (k*rh) ]
  
  filtblocks <- filterScreenRect tlwidget 
  forM_ filtblocks $ \(orect1,rect1,_,b) -> do
    let (rf,gf,bf) = _block_color b
        cf a = round (a*255.0) :: Int
        (red,green,blue) = (cf rf , cf gf , cf bf)  
        col = colorRGB red green blue
        textcol = if (red + green + blue) > 3*128 then colorRGB 0 0 0 else colorRGB 255 255 255 
        brush = brushSolid col
        Rect x1 y1 w1 h1 = fromRect1 rect1
        Rect x0 y0 w0 h0 = fromRect1 orect1
        rh = _rowheight tlwidget
    dcWithBrushStyle dc brush $ 
      --dcWithPenStyle dc penTransparent $ do
        dcDrawRectangle dc (Rect (x1) (y1-1) (w1+1) (h1+2))
    case Just (_block_name b) of 
      Nothing -> return ()
      Just txt -> do
        dcWithFontStyle dc fontDefault $ do
          Size xs ys <- getTextExtent dc txt    -- getTextExtent needs the fontstyle
          Control.Monad.when (w0 > xs + 5 ) $ do 
            dcSetTextForeground dc textcol 
            dcDrawText dc txt (Point (x0+4) (y0 + 1 + (rh-ys) `div` 2))


  return ()
       
-------------------------------------------------------
